/*
 * Copyright (c) 2002, 2009 Jens Keiner, Stefan Kunis, Daniel Potts
 * Copyright (c) 2012-2013 Michael Pippig
 *
 * This file is part of PNFFT.
 *
 * PNFFT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PNFFT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PNFFT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#include "pnfft.h"
#include "ipnfft.h"
#include "bspline.h"

static void bspline_help(
    int k, R x, R *scratch, int j,
    int ug, int og, int r
    )
{
  int i;   /* row index of the de Boor scheme */
  int idx; /* index in scratch */
  R a;     /* alpha of the de Boor scheme */

  /* computation of one column */
  for (i = og + r - k + 1, idx = og; idx >= ug; i--, idx--){
    a = ((R)(x - i)) / ((R)(k - j));
    scratch[idx] = (1 - a) * scratch[idx-1] + a * scratch[idx];
  }
} /* bspline_help */

/** Computes \f$M_{k,0}\left(x\right)\f$
 *  scratch is used for de Boor's scheme
 */
R PNX(bspline)(
    int k, R x, R *scratch
    )
{
  R result_value;                  /**< M_{k,0}\left(x\right)            */
  int r;                           /**< \f$x \in {\rm supp}(M_{0,r}) \f$ */
  int g1,g2;                       /**< boundaries                       */
  int j,idx,ug,og;                 /**< indices                          */
  R a;                             /**< alpha of the de Boor scheme      */

  result_value=0.0;
  if(0<x && x<k){
    /* using symmetry around k/2 */
    if((k-x)<x) x=k-x;

    r=(int)(pnfft_ceil(x)-1.0);

    for(idx=0; idx<k; idx++)
      scratch[idx]=0.0;
    scratch[k-r-1]=1.0;

    /* bounds of the algorithm */
    g1 = r;
    g2 = k - 1 - r;
    ug = g2;

    /* g1<=g2 holds */
    for(j=1, og=g2+1; j<=g1; j++, og++){
      a = (x - r + k - 1 - og)/(k - j);
      scratch[og] = (1 - a) * scratch[og-1];
      bspline_help(k,x,scratch,j,ug+1,og-1,r);
      a = (x - r + k - 1 - ug)/(k - j);
      scratch[ug] = a * scratch[ug];
    }
    for(og-- ; j<=g2; j++){
      bspline_help(k,x,scratch,j,ug+1,og,r);
      a = (x - r + k - 1 - ug)/(k - j);
      scratch[ug] = a * scratch[ug];
    }
    for( ; j<k; j++){
      ug++;
      bspline_help(k,x,scratch,j,ug,og,r);
    }
    result_value = scratch[k-1];
  }
  return(result_value);
} /* bspline */


