/*
 * Copyright (c) 2013 Michael Pippig
 *
 * This file is part of PNFFT.
 *
 * PNFFT is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PNFFT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PNFFT.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

/* The matrix B is very similar for the interlaced and non-interlaced NFFT.
 * We do not want to duplicate a lot of code, therefore we include this file
 * two times into ndft-parallel.c with PNFFT_ENABLE_INTERLACING on/off.*/


#if PNFFT_ENABLE_INTERLACING
void PNX(trafo_B_interlaced)(
    PNX(plan) ths
    )
#else
void PNX(trafo_B)(
    PNX(plan) ths
    )
#endif
{
  const int cutoff = ths->cutoff;
  INT j, l0, l1, l2, m0, m1, m2;
  INT *sorted_index = NULL;
  INT local_no[3], local_no_start[3];
  INT gcells_below[3], gcells_above[3];
  INT u_j[3], local_ngc[3];
  R *pre_psi = NULL;
  R *pre_dpsi = NULL;
 
  local_size_B(ths,
      local_no, local_no_start);

#if PNFFT_ENABLE_DEBUG
  int myrank;
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  R rsum, grsum;
  R rsum_derive, grsum_derive;
  C csum, gcsum;
  
  if(!myrank) fprintf(stderr, "no = [%td, %td, %td]\n", ths->no[0], ths->no[1], ths->no[2]);
//  if(!myrank) fprintf(stderr, "local_no = [%td, %td, %td], local_ngc = [%td, %td, %td], local_ngc_total = %td\n",
//      local_no[0], local_no[1], local_no[2], local_ngc[0], local_ngc[1], local_ngc[2], local_ngc_total);
      
  csum = 0.0;
  for(INT t=0; t<local_no[0]*local_no[1]*local_no[2]; t++)
    csum += pnfft_fabs(pnfft_creal(ths->g2[t])) + _Complex_I * pnfft_fabs(pnfft_cimag(ths->g2[t])) ;
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of Fourier coefficients before twiddles: %e + I* %e\n", pnfft_creal(gcsum), pnfft_cimag(gcsum));
#endif

  /* perform fftshift */
  ths->timer_trafo[PNFFT_TIMER_SHIFT_INPUT] -= MPI_Wtime();
//   if(ths->pnfft_flags & PNFFT_SHIFTED_IN)
    fftshift_N_half(ths->N, ths->n, local_no, local_no_start, PFFT_FORWARD,
        ths->g2);
  ths->timer_trafo[PNFFT_TIMER_SHIFT_INPUT] += MPI_Wtime();

  get_size_gcells(ths->m, ths->cutoff, ths->pnfft_flags,
      gcells_below, gcells_above);
  local_array_size(local_no, gcells_below, gcells_above,
      local_ngc);

#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(INT t=0; t<local_no[0]*local_no[1]*local_no[2]; t++)
    csum += pnfft_fabs(pnfft_creal(ths->g2[t])) + _Complex_I * pnfft_fabs(pnfft_cimag(ths->g2[t])) ;
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of Fourier coefficients before ghostcell send: %e + I* %e\n", pnfft_creal(gcsum), pnfft_cimag(gcsum));
#endif

  /* send ghost cells in ring */
  ths->timer_trafo[PNFFT_TIMER_GCELLS] -= MPI_Wtime();
  PX(exchange)(ths->gcplan);
  ths->timer_trafo[PNFFT_TIMER_GCELLS] += MPI_Wtime();

#if PNFFT_ENABLE_DEBUG
  INT local_ngc_total = PNX(prod_INT)(3, local_ngc);
#endif

#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(INT t=0; t<local_ngc_total; t++)
    csum += pnfft_fabs(pnfft_creal(ths->g2[t])) + _Complex_I * pnfft_fabs(pnfft_cimag(ths->g2[t])) ;
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of Fourier coefficients after ghostcell send: %e + I* %e\n", pnfft_creal(gcsum), pnfft_cimag(gcsum));
#endif  

#if PNFFT_ENABLE_DEBUG
  rsum = 0.0;
  for(INT j=0; j<3*ths->local_M; j++)
    rsum += pnfft_fabs(ths->x[j]);
  MPI_Reduce(&rsum, &grsum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of x before sort: %e\n", grsum);
#endif

  /* sort indices for better cache handling */
  if(ths->pnfft_flags & PNFFT_SORT_NODES){
    ths->timer_trafo[PNFFT_TIMER_SORT_NODES] -= MPI_Wtime();
    sorted_index = (INT*) PNX(malloc)(sizeof(INT) * (size_t) 2*ths->local_M);
    sort_nodes_for_better_cache_handle(
        ths->d, ths->n, ths->m, ths->local_M, ths->x,
        sorted_index);
    ths->timer_trafo[PNFFT_TIMER_SORT_NODES] += MPI_Wtime();
  }

#if PNFFT_ENABLE_DEBUG
  rsum = 0.0;
  for(j=0; j<3*ths->local_M; j++)
    rsum += pnfft_fabs(ths->x[j]);
  MPI_Reduce(&rsum, &grsum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of x after sort: %e\n", grsum);
#endif

#if PNFFT_ENABLE_DEBUG
  rsum = 0.0;
#endif
#if PNFFT_ENABLE_DEBUG
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F)
    rsum_derive = 0.0;
#endif

  ths->timer_trafo[PNFFT_TIMER_LOOP_B] -= MPI_Wtime();
  pre_psi = (R*) PNX(malloc)(sizeof(R) * (size_t) cutoff*3);
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F)
    pre_dpsi = (R*) PNX(malloc)(sizeof(R) * (size_t) cutoff*3);
  for(INT p=0; p<ths->local_M; p++){
    j = (ths->pnfft_flags & PNFFT_SORT_NODES) ? sorted_index[2*p+1] : p;
    
    /* evaluate window on axes */
    if(ths->pnfft_flags & PNFFT_PRE_INTPOL_PSI)
      pre_psi_tensor_intpol(
          ths->n, ths->m, ths->cutoff, &(ths->x[ths->d*j]),
          ths->intpol_order, ths->intpol_num_nodes, ths->intpol_tables_psi,
          pre_psi);
    else
      pre_psi_tensor(
          ths->n, ths->b, ths->m, ths->cutoff, &(ths->x[ths->d*j]),
          ths->exp_const, ths->spline_coeffs, ths->pnfft_flags,
          pre_psi);
  
#if PNFFT_ENABLE_DEBUG
    for(int t=0; t<3*cutoff; t++)
      rsum += pnfft_fabs(pre_psi[t]);
#endif
  
    if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
      if(ths->pnfft_flags & PNFFT_PRE_INTPOL_PSI)
        pre_dpsi_tensor_intpol(
            ths->n, ths->m, ths->cutoff, &(ths->x[ths->d*j]),
            ths->intpol_order, ths->intpol_num_nodes, ths->intpol_tables_dpsi,
            pre_dpsi);
      else
        pre_dpsi_tensor(
            ths->n, ths->b, ths->m, ths->cutoff, &(ths->x[ths->d*j]), ths->spline_coeffs,
            pre_psi, ths->pnfft_flags,
            pre_dpsi);
    }

#if PNFFT_ENABLE_DEBUG
    if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F)
      for(int t=0; t<3*cutoff; t++)
        rsum_derive += pnfft_fabs(pre_dpsi[t]);
#endif

    lowest_summation_index(
        ths->n, ths->m, &(ths->x[ths->d*j]), local_no_start, gcells_below,
        u_j);
    
    /* compute f */
    if(ths->compute_flags & PNFFT_COMPUTE_F){
      ths->f[j] = 0;
      m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
      for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2])
        for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2])
          for(l2=0, m2 = m1; l2<cutoff; l2++, m2++)
#if PNFFT_ENABLE_INTERLACING
            ths->f[j] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * pnfft_creal(ths->g2[m2]);
#else
            ths->f[j] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->g2[m2];
#endif
    }

    /* compute grad_f */
    if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
      for(int t=0; t<ths->d; t++)
        ths->grad_f[ths->d*j+t] = 0;
      m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
      for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2]){
        for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2]){
          for(l2=0, m2 = m1; l2<cutoff; l2++, m2++){
#if PNFFT_ENABLE_INTERLACING
            ths->grad_f[ths->d*j+0] += pre_dpsi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * pnfft_creal(ths->g2[m2]);
            ths->grad_f[ths->d*j+1] += pre_psi[l0] * pre_dpsi[cutoff+l1] * pre_psi[2*cutoff+l2] * pnfft_creal(ths->g2[m2]);
            ths->grad_f[ths->d*j+2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_dpsi[2*cutoff+l2] * pnfft_creal(ths->g2[m2]);
#else
            ths->grad_f[ths->d*j+0] += pre_dpsi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->g2[m2];
            ths->grad_f[ths->d*j+1] += pre_psi[l0] * pre_dpsi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->g2[m2];
            ths->grad_f[ths->d*j+2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_dpsi[2*cutoff+l2] * ths->g2[m2];
#endif
          }
        }
      }
    }

#if PNFFT_ENABLE_INTERLACING
    R x[3];
    
    /* shift x by half the mesh width */
    for(int t=0; t<3; t++)
      x[t] = ths->x[ths->d*j+t] + 0.5/ths->n[t];

    /* We need to compute the lowest summation index before we fold x back into [-0.5,0.5).
     * Otherwise u_j may be also folded and gets less than the local offset local_no_start. */
    lowest_summation_index(
        ths->n, ths->m, x, local_no_start, gcells_below,
        u_j);
    
    /* assure -0.5 <= x < 0.5 */
    for(int t=0; t<3; t++)
      if(x[t] >= 0.5)
        x[t] -= 1.0;

    /* evaluate window on axes */
    if(ths->pnfft_flags & PNFFT_PRE_INTPOL_PSI)
      pre_psi_tensor_intpol(
          ths->n, ths->m, ths->cutoff, x,
          ths->intpol_order, ths->intpol_num_nodes, ths->intpol_tables_psi,
          pre_psi);
    else
      pre_psi_tensor(
          ths->n, ths->b, ths->m, ths->cutoff, x,
          ths->exp_const, ths->spline_coeffs, ths->pnfft_flags,
          pre_psi);

    if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
      if(ths->pnfft_flags & PNFFT_PRE_INTPOL_PSI)
        pre_dpsi_tensor_intpol(
            ths->n, ths->m, ths->cutoff, x,
            ths->intpol_order, ths->intpol_num_nodes, ths->intpol_tables_dpsi,
            pre_dpsi);
      else
        pre_dpsi_tensor(
            ths->n, ths->b, ths->m, ths->cutoff, x, ths->spline_coeffs,
            pre_psi, ths->pnfft_flags,
            pre_dpsi);
    }
    
    /* compute f with shifted x */
    if(ths->compute_flags & PNFFT_COMPUTE_F){
      m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
      for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2])
        for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2])
          for(l2=0, m2 = m1; l2<cutoff; l2++, m2++)
            ths->f[j] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * pnfft_cimag(ths->g2[m2]);

      ths->f[j] /= 2.0;
    }

    /* compute grad_f with shifted x */
    if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
      m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
      for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2]){
        for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2]){
          for(l2=0, m2 = m1; l2<cutoff; l2++, m2++){
            ths->grad_f[ths->d*j+0] += pre_dpsi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * pnfft_cimag(ths->g2[m2]);
            ths->grad_f[ths->d*j+1] += pre_psi[l0] * pre_dpsi[cutoff+l1] * pre_psi[2*cutoff+l2] * pnfft_cimag(ths->g2[m2]);
            ths->grad_f[ths->d*j+2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_dpsi[2*cutoff+l2] * pnfft_cimag(ths->g2[m2]);
          }
        }
      }
      for(int t=0; t<ths->d; t++)
        ths->grad_f[ths->d*j+t] /= 2.0;
    }
#endif
  }

  ths->timer_trafo[PNFFT_TIMER_LOOP_B] += MPI_Wtime();

#if PNFFT_ENABLE_DEBUG
  MPI_Reduce(&rsum, &grsum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of pre_psi: %e\n", grsum);
#endif  

#if PNFFT_ENABLE_DEBUG
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
    MPI_Reduce(&rsum_derive, &grsum_derive, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    if(!myrank) fprintf(stderr, "PNFFT: Sum of pre_dpsi: %e\n", grsum_derive);
  }
#endif  

#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(j=0; j<ths->local_M; j++)
    csum += pnfft_fabs(pnfft_creal(ths->f[j])) + _Complex_I * pnfft_fabs(pnfft_cimag(ths->f[j]));
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT: Sum of f: %e + I* %e\n", pnfft_creal(gcsum), pnfft_cimag(gcsum));
#endif
#if PNFFT_ENABLE_DEBUG
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
    csum = 0.0;
    for(j=0; j<ths->local_M; j++)
      csum += pnfft_fabs(pnfft_creal(ths->grad_f[3*j+0])) + _Complex_I * pnfft_fabs(pnfft_cimag(ths->grad_f[3*j+0]));
    MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    if(!myrank) fprintf(stderr, "PNFFT: Sum of 1st component of grad_f: %e + I* %e\n", pnfft_creal(gcsum), pnfft_cimag(gcsum));
  }
#endif
#if PNFFT_ENABLE_DEBUG
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
    csum = 0.0;
    for(j=0; j<ths->local_M; j++)
      csum += pnfft_fabs(pnfft_creal(ths->grad_f[3*j+1])) + _Complex_I * pnfft_fabs(pnfft_cimag(ths->grad_f[3*j+1]));
    MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    if(!myrank) fprintf(stderr, "PNFFT: Sum of 2nd component of grad_f: %e + I* %e\n", pnfft_creal(gcsum), pnfft_cimag(gcsum));
  }
#endif
#if PNFFT_ENABLE_DEBUG
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F){
    csum = 0.0;
    for(j=0; j<ths->local_M; j++)
      csum += pnfft_fabs(pnfft_creal(ths->grad_f[3*j+2])) + _Complex_I * pnfft_fabs(pnfft_cimag(ths->grad_f[3*j+2]));
    MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
    if(!myrank) fprintf(stderr, "PNFFT: Sum of 3rd component of grad_f: %e + I* %e\n", pnfft_creal(gcsum), pnfft_cimag(gcsum));
  }
#endif
  
  if(ths->pnfft_flags & PNFFT_SORT_NODES)
    PNX(free)(sorted_index);

  PNX(free)(pre_psi);
  if(ths->compute_flags & PNFFT_COMPUTE_GRAD_F)
    PNX(free)(pre_dpsi);
}


#if PNFFT_ENABLE_INTERLACING
void PNX(adjoint_B_interlaced)(
    PNX(plan) ths
    )
#else
void PNX(adjoint_B)(
    PNX(plan) ths
    )
#endif
{
  const int cutoff = ths->cutoff;
  INT j, l0, l1, l2, m0, m1, m2;
  INT *sorted_index = NULL;
  INT local_no[3], local_no_start[3];
  INT gcells_below[3], gcells_above[3];
  INT u_j[3], local_ngc[3], local_ngc_total;
  R *pre_psi = NULL;
  
  local_size_B(ths,
      local_no, local_no_start);

  get_size_gcells(ths->m, ths->cutoff, ths->pnfft_flags,
      gcells_below, gcells_above);
  local_array_size(local_no, gcells_below, gcells_above,
      local_ngc);

  local_ngc_total = PNX(prod_INT)(3, local_ngc);
  memset(ths->g2, 0, local_ngc_total * sizeof(C));

#if PNFFT_ENABLE_DEBUG
  int myrank;
  MPI_Comm_rank(MPI_COMM_WORLD, &myrank);
  R rsum, grsum;
  C csum, gcsum;
  
  if(!myrank) fprintf(stderr, "no = [%td, %td, %td]\n", ths->no[0], ths->no[1], ths->no[2]);
//  if(!myrank) fprintf(stderr, "local_no = [%td, %td, %td], local_ngc = [%td, %td, %td], local_ngc_total = %td\n",
//      local_no[0], local_no[1], local_no[2], local_ngc[0], local_ngc[1], local_ngc[2], local_ngc_total);
      
  rsum = 0.0;
  for(INT j=0; j<3*ths->local_M; j++)
    rsum += pnfft_fabs(ths->x[j]);
  MPI_Reduce(&rsum, &grsum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of x before sort: %e\n", grsum);
#endif

  /* sort indices for better cache handling */
  if(ths->pnfft_flags & PNFFT_SORT_NODES){
    ths->timer_adj[PNFFT_TIMER_SORT_NODES] -= MPI_Wtime();
    sorted_index = (INT*) PNX(malloc)(sizeof(INT) * (size_t) 2*ths->local_M);
    sort_nodes_for_better_cache_handle(
        ths->d, ths->n, ths->m, ths->local_M, ths->x,
        sorted_index);
    ths->timer_adj[PNFFT_TIMER_SORT_NODES] += MPI_Wtime();
  }
  
#if PNFFT_ENABLE_DEBUG
  rsum = 0.0;
  for(j=0; j<3*ths->local_M; j++)
    rsum += pnfft_fabs(ths->x[j]);
  MPI_Reduce(&rsum, &grsum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of x after sort: %e\n", grsum);
#endif
  
#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(j=0; j<ths->local_M; j++)
    csum += pnfft_fabs(pnfft_creal(ths->f[j])) + _Complex_I * pnfft_fabs(pnfft_cimag(ths->f[j]));
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of f: %e + I* %e\n", pnfft_creal(gcsum), pnfft_cimag(gcsum));
#endif
  
#if PNFFT_ENABLE_DEBUG
  rsum = 0.0;
#endif

  ths->timer_adj[PNFFT_TIMER_LOOP_B] -= MPI_Wtime();
  pre_psi = (R*) PNX(malloc)(sizeof(R) * (size_t) cutoff*3);
  for(INT p=0; p<ths->local_M; p++){
    j = (ths->pnfft_flags & PNFFT_SORT_NODES) ? sorted_index[2*p+1] : p;

    /* evaluate window on axes */
    if(ths->pnfft_flags & PNFFT_PRE_INTPOL_PSI)
      pre_psi_tensor_intpol(
          ths->n, ths->m, ths->cutoff, &(ths->x[ths->d*j]),
          ths->intpol_order, ths->intpol_num_nodes, ths->intpol_tables_psi,
          pre_psi);
    else
      pre_psi_tensor(
          ths->n, ths->b, ths->m, ths->cutoff, &(ths->x[ths->d*j]),
          ths->exp_const, ths->spline_coeffs, ths->pnfft_flags,
          pre_psi);

#if PNFFT_ENABLE_DEBUG
    for(int t=0; t<3*cutoff; t++)
      rsum += pnfft_fabs(pre_psi[t]);
#endif

    lowest_summation_index(
        ths->n, ths->m, &(ths->x[ths->d*j]), local_no_start, gcells_below,
        u_j);
    
#if PNFFT_ENABLE_DEBUG
    R lo[3], up[3];
    PNX(node_borders)(ths->n, local_no, local_no_start, ths->x_max,
        lo, up);

    int failed=0;
    m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
    for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2]){
      for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2]){
        for(l2=0, m2 = m1; l2<cutoff; l2++, m2++ ){
           if((m2 < 0) && (!failed)){
             fprintf(stderr, "!!! Error: myrank = %d, write to g2 out of range (m2<0), m2=%td !!!\n", myrank, m2);
             failed = 1;
           }
           if((m2 >= local_ngc_total) && (!failed)){
             fprintf(stderr, "!!! Error: write to g2 out of range (m2>=local_ngc_total)i, m2=%td, local_ngc_total=%td !!!\n", m2, local_ngc_total);
             failed = 1;
           }
           ths->g2[m2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->f[j];
        }
      }
    }
    if(failed){
      fprintf(stderr, "myrank = %d, u_j=[%td,%td,%td], m0 = %td !!!\n", myrank, u_j[0], u_j[1], u_j[2], m0);
      fprintf(stderr, "local_M = %td, local_no = [%td, %td, %td]\n",
      ths->local_M, local_no[0], local_no[1], local_no[2]);
      fprintf(stderr, "x = [%f, %f, %f]i, lower_border = [%f, %f, %f], upper_border = [%f, %f, %f]\n",
      ths->x[ths->d*j], ths->x[ths->d*j+1], ths->x[ths->d*j+2],
        lo[0], lo[1], lo[2], up[0], up[1], up[2]);
    }
#else
    m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
    for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2])
      for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2])
        for(l2=0, m2 = m1; l2<cutoff; l2++, m2++ )
#if PNFFT_ENABLE_INTERLACING
           ths->g2[m2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * pnfft_creal(ths->f[j]);
#else
           ths->g2[m2] += pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * ths->f[j];
#endif
#endif
#if PNFFT_ENABLE_INTERLACING
    R x[3];

    /* shift x by half the mesh width */
    for(int t=0; t<3; t++)
      x[t] = ths->x[ths->d*j+t] + 0.5/ths->n[t];

    /* We need to compute the lowest summation index before we fold x back into [-0.5,0.5).
     * Otherwise u_j may be also folded and gets less than the local offset local_no_start. */
    lowest_summation_index(
        ths->n, ths->m, x, local_no_start, gcells_below,
        u_j);

    /* assure -0.5 <= x < 0.5 */
    for(int t=0; t<3; t++)
      if(x[t] >= 0.5)
        x[t] -= 1.0;

    /* evaluate window on axes */
    if(ths->pnfft_flags & PNFFT_PRE_INTPOL_PSI)
      pre_psi_tensor_intpol(
          ths->n, ths->m, ths->cutoff, x,
          ths->intpol_order, ths->intpol_num_nodes, ths->intpol_tables_psi,
          pre_psi);
    else
      pre_psi_tensor(
          ths->n, ths->b, ths->m, ths->cutoff, x,
          ths->exp_const, ths->spline_coeffs, ths->pnfft_flags,
          pre_psi);

    m0 = PNFFT_PLAIN_INDEX_3D(u_j, local_ngc);
    for(l0=0; l0<cutoff; l0++, m0 += local_ngc[1]*local_ngc[2])
      for(l1=0, m1=m0; l1<cutoff; l1++, m1 += local_ngc[2])
        for(l2=0, m2 = m1; l2<cutoff; l2++, m2++ )
           ths->g2[m2] += I * pre_psi[l0] * pre_psi[cutoff+l1] * pre_psi[2*cutoff+l2] * pnfft_creal(ths->f[j]);
#endif
  }
  ths->timer_adj[PNFFT_TIMER_LOOP_B] += MPI_Wtime();

#if PNFFT_ENABLE_DEBUG
  MPI_Reduce(&rsum, &grsum, 1, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of pre_psi: %e\n", grsum);
#endif  

#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(INT t=0; t<local_ngc_total; t++)
    csum += pnfft_fabs(pnfft_creal(ths->g2[t])) + _Complex_I * pnfft_fabs(pnfft_cimag(ths->g2[t])) ;
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of Fourier coefficients before ghostcell reduce: %e + I* %e\n", pnfft_creal(gcsum), pnfft_cimag(gcsum));
#endif  

  /* reduce ghost cells in ring */
  ths->timer_adj[PNFFT_TIMER_GCELLS] -= MPI_Wtime();
  PX(reduce)(ths->gcplan);
  ths->timer_adj[PNFFT_TIMER_GCELLS] += MPI_Wtime();

#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(INT t=0; t<local_no[0]*local_no[1]*local_no[2]; t++)
    csum += pnfft_fabs(pnfft_creal(ths->g2[t])) + _Complex_I * pnfft_fabs(pnfft_cimag(ths->g2[t])) ;
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of Fourier coefficients after ghostcell reduce: %e + I* %e\n", pnfft_creal(gcsum), pnfft_cimag(gcsum));
#endif
 
  /* perform fftshift */
  ths->timer_adj[PNFFT_TIMER_SHIFT_INPUT] -= MPI_Wtime();
//   if(ths->pnfft_flags & PNFFT_SHIFTED_IN)
    fftshift_N_half(ths->N, ths->n, local_no, local_no_start, PFFT_BACKWARD,
        ths->g2);
  ths->timer_adj[PNFFT_TIMER_SHIFT_INPUT] += MPI_Wtime();

#if PNFFT_ENABLE_DEBUG
  csum = 0.0;
  for(INT t=0; t<local_no[0]*local_no[1]*local_no[2]; t++)
    csum += pnfft_fabs(pnfft_creal(ths->g2[t])) + _Complex_I * pnfft_fabs(pnfft_cimag(ths->g2[t])) ;
  MPI_Reduce(&csum, &gcsum, 2, MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD);
  if(!myrank) fprintf(stderr, "PNFFT^H: Sum of Fourier coefficients after twiddles: %e + I* %e\n", pnfft_creal(gcsum), pnfft_cimag(gcsum));
#endif
  
  if(ths->pnfft_flags & PNFFT_SORT_NODES)
    PNX(free)(sorted_index);
  PNX(free)(pre_psi);
}


